#ifndef CAMERA_H
#define CAMERA_H

#include "math.h"

#include <QMatrix4x4>
#include <QVector2D>
#include <QVector3D>

/**
 * @brief Class for orbital cameras.
 */
class Camera : public QObject
{
	Q_OBJECT

	public:

		/**
		 * @brief Constructs a new camera.
		 * @param parent the parent object.
		 * @param minDist the minimum distance from the target.
		 * @param maxDist the maximum distance from the target.
		 * @param initDist the initial distance.
		 * @param initYaw the initial yaw angle.
		 * @param initPitch the initial pitch angle
		 */
		Camera(QObject* parent, float minDist, float maxDist, float initDist = .0f, float initYaw = .0f, float initPitch = .0f);

		/**
		 * @brief Constructs the view matrix and returns it in \p viewMatrix.
		 * @param viewMatrix returns the view matrix.
		 */
		void getViewMatrix(QMatrix4x4& viewMatrix);

		/**
		 * @brief Constructs the projection matrix and returns it in \p projMatrix.
		 * @param projMatrix returns the projection matrix.
		 */
		void getProjectionMatrix(QMatrix4x4& projMatrix);

	public slots:

		/**
		 * @brief Sets the aspect ratio for the perspective transform.
		 * @param aspectRatio the aspect ratio.
		 */
		void setAspectRatio(float aspectRatio);

		/**
		 * @brief Sets the z values for the near and far plane respectively.
		 * @param near the near plane.
		 * @param far the far plane.
		 */
		void setNearFarPlane(float near, float far);

		/**
		 * @brief Yaws the camera
		 * @param angle in radians
		 */
		void rotateX(float angle);

		/**
		 * @brief Pitches the camera
		 * @param angle in radians
		 */
		void rotateY(float angle);

		/**
		 * @brief Offets the camera.
		 * @param offset the offset vector
		 */
		void move(const QVector2D& offset);

		/**
		 * @brief Changes the distance of the camera
		 * @param distance the distance value
		 */
		void zoom(float distance);

	private:

		// aspect ratio
		float m_aspectRatio;

		// near plane
		float m_near;

		// far plane
		float m_far;

		// current camera view distance
		float m_dist;

		// minimal camera view distance
		float m_minDist;

		// maximal camera view distance
		float m_maxDist;

		// current camera yaw angle in radians
		float m_xRot;

		//current camera pitch angle in radians
		float m_yRot;

		// camera x/y offset
		QVector2D m_offset;
};

#endif // CAMERA_H
