#include "glwidget.h"

#include <QMouseEvent>
#include <QDebug>
#include <QtMath>

GLWidget::GLWidget(QWidget* parent, int updateRate, bool sampleShading, int anisotropy)
	: QOpenGLWidget(parent), sampleShading(sampleShading), anisotropy(anisotropy)
{
	setFocusPolicy(Qt::StrongFocus);

	// Update widget and scene at a fixed interval.
	// This won't work flawlessly, since QTimer only
	// offers limited precision. So the desired update rate will not
	// match the actual rate in some cases. Also the refresh rate of
	// context seems to be capped at 60.
	connect(&updateTimer, &QTimer::timeout, [this] () {
		updateFrame();
		update();
	});

	updateTimer.setTimerType(Qt::PreciseTimer);
	updateTimer.start(1000 / qMax(updateRate, 1));

	// Every second, signal fps and restart counter
	connect(&fpsTimer, &QTimer::timeout, [this] () {
		emit frameRate(frameCounter);
		frameCounter = 0;
	});

	fpsTimer.setTimerType(Qt::PreciseTimer);
	fpsTimer.start(1000);

	// Start timer to compute delta time
	frameTimer.start();

	// Camera
    camera = new Camera(this, 5, 120, 90, Math::PI / 2, Math::PI / 2); //FoV based
}

GLWidget::~GLWidget()
{
	delete textureStroke;
    delete textureStripMask;
}

void GLWidget::initializeGL()
{
	// Setup OpenGL functions & print version
	if (!initializeOpenGLFunctions())
	{
		qDebug() << "Failed to initialize OpenGL functions.";
	}
	qDebug("Using OpenGL %s", (const char*) glGetString(GL_VERSION));

	// Setup OpenGL logger
	logger = new QOpenGLDebugLogger(this);

	if (context()->hasExtension(QByteArrayLiteral("GL_KHR_debug")) && logger->initialize())
	{
		connect(logger, &QOpenGLDebugLogger::messageLogged, [] (const QOpenGLDebugMessage& m) {
			qDebug() << qPrintable(m.message());
		});

		logger->startLogging();
		logger->enableMessages();
	}
	else
	{
		qWarning() << "OpenGL debugger not available!";
	}

	// Load stroke texture
	const QString pathStroke("textures/stroke.png");

	QImage imgStroke(pathStroke);
	if (imgStroke.isNull())
	{
		qDebug("Failed to load texture '%s'", qPrintable(pathStroke));
	}

	textureStroke = new QOpenGLTexture(imgStroke);
	textureStroke->setMinificationFilter(QOpenGLTexture::LinearMipMapLinear);
	textureStroke->setMagnificationFilter(QOpenGLTexture::Linear);
    textureStroke->setWrapMode(QOpenGLTexture::ClampToEdge);
	textureStroke->setMaximumAnisotropy(anisotropy);

    // Load strip texture
    const QString pathStrip("textures/stripMask.png");

    QImage imgStrip(pathStrip);
    if (imgStrip.isNull())
    {
        qDebug("Failed to load texture '%s'", qPrintable(pathStrip));
    }

    textureStripMask = new QOpenGLTexture(imgStrip);
    textureStripMask->setMinificationFilter(QOpenGLTexture::LinearMipMapLinear);
    textureStripMask->setMagnificationFilter(QOpenGLTexture::Linear);
	textureStripMask->setWrapMode(QOpenGLTexture::ClampToEdge);
	textureStripMask->setMaximumAnisotropy(anisotropy);

	// Create shader program
	program = createShaderProgram("shader/simple.vert", "shader/simple.frag");
    program->bind();

    // Create VAO and bind it
    if (!vao.create())
    {
        qDebug() << "Failed to create VAO.";
    }
    QOpenGLVertexArrayObject::Binder vaoBinder(&vao);


	// Vertex buffer
	if (!vbo.create())
	{
		qDebug() << "Failed to create VBO";
	}

	vbo.bind();
	vbo.setUsagePattern(QOpenGLBuffer::StaticDraw);
    program->enableAttributeArray(0);
    program->setAttributeBuffer(0, GL_FLOAT, 0, 3, 0);

    // TexCoord buffer
    if (!uvBuffer.create())
    {
        qDebug() << "Failed to create UV VBO";
    }

    uvBuffer.bind();
    uvBuffer.setUsagePattern(QOpenGLBuffer::StaticDraw);
    program->enableAttributeArray(1);
	program->setAttributeBuffer(1, GL_FLOAT, 0, 2, 0);

    // Direction buffer
    if (!dirBuffer.create())
    {
        qDebug() << "Failed to create VBO";
    }

    dirBuffer.bind();
    dirBuffer.setUsagePattern(QOpenGLBuffer::StaticDraw);
    program->enableAttributeArray(2);
    program->setAttributeBuffer(2, GL_FLOAT, 0, 3, 0);

    program->release();

	vbo.release();
    uvBuffer.release();
	dirBuffer.release();

	// Sample shading
	if (sampleShading)
	{
		glEnable(GL_SAMPLE_SHADING);
		glMinSampleShading(0.25f);
	}

	// Setup render states etc.
	glEnable(GL_DEPTH_TEST);
	glClearColor(1, 1, 1, 1);
}

QOpenGLShaderProgram* GLWidget::createShaderProgram(const QString &pathVS, const QString &pathFS, const QString &pathGS)
{
	// Create shader program
	QOpenGLShaderProgram* p = new QOpenGLShaderProgram(this);
	if (!p->create())
	{
		qDebug() << "Failed to create shader program";
		return Q_NULLPTR;
	}

	// Add shaders
	p->addShaderFromSourceFile(QOpenGLShader::Vertex, pathVS);
	p->addShaderFromSourceFile(QOpenGLShader::Fragment, pathFS);

	if (!pathGS.isEmpty())
	{
		p->addShaderFromSourceFile(QOpenGLShader::Geometry, pathGS);
	}

	// Link
	return p->link() ? p : Q_NULLPTR;
}

void GLWidget::setTrackData(const QVector<Track>&& tracks)
{
	if (tracks.empty())
	{
		return;
	}

	this->tracks = tracks;

	// Compute range of data
	const float inf = std::numeric_limits<float>::infinity();
	QVector3D min = QVector3D(inf, inf, inf);
	QVector3D max = QVector3D(-inf, -inf, -inf);

	foreach (const Track& t, this->tracks)
	{
		foreach (const QVector3D& v, t)
		{
			min.setX(qMin(min.x(), v.x()));
			min.setY(qMin(min.y(), v.y()));
			min.setZ(qMin(min.z(), v.z()));

			max.setX(qMax(max.x(), v.x()));
			max.setY(qMax(max.y(), v.y()));
			max.setZ(qMax(max.z(), v.z()));
		}
	}

    //printf("min = %.3f/%.3f/%.3f, max = %.3f/%.3f/%.3f\n", min.x(), min.y(), min.z(), max.x(), max.y(), max.z());

	float maxRange = qMax(qMax((max.x() - min.x()), (max.y() - min.y())), (max.z()-min.z()));
	float scale = 2.0f/maxRange;
	QVector3D range = QVector3D(max-min);

	// Compute total number of vertices & filter
	QVector<Track> filteredTracks;

	numVertices = 0;
	int numPoints = 0;

	foreach (const Track& t, this->tracks)
	{
		Track tf;

		bool zNeg = false, zPos = false;

		foreach (const QVector3D& v, t)
		{
			QVector3D vn = normalizedPoint(v, min, range, scale);
			tf.append(vn);

			zNeg = zNeg || (vn.z() < 0);
			zPos = zPos || (vn.z() > 0);
		}

		if (!filterTracks || (zPos && zNeg))
		{
			numPoints += t.size();
			numVertices += 2 * (t.size() - 1);	// n - 1 segments

			filteredTracks.append(tf);
		}
	}

	// Copy data to CPU buffer
	QVector<float> vertices;
    QVector<float> UVs;
    QVector<float> directions;

    vertices.reserve(numVertices * 9); //6 Positions per line segment
	UVs.reserve(numVertices * 6);
	directions.reserve(numVertices * 9);

    //Tapering
    int nt = 2;
    float tt = 0.2f;

	foreach (const Track& t, filteredTracks)
	{
		for (int i = 0; i < t.size() - 1; i++)
		{
			QVector3D p1 = t[i];
			QVector3D p2 = t[i + 1];

            float u1;// = (float)i / (float)t.size();
            float u2;// = (float)(i+1) / (float)t.size();

            //Tapering
            if (t.size() >= nt*2) {
                //printf("size=%d > nt=%d\n", t.size(), nt);
                if (i < nt) {
                    u1 = ((float)i / (float)nt) * tt;
                    u2 = ((float)(i+1) / (float)nt) * tt;
                }else if (i > t.size()-(2+nt)) {
                    u1 = (1.0f-tt) + ((float)(i-(t.size()-(1+nt))) / (float)nt) * tt;
                    u2 = (1.0f-tt) + ((float)(i+1-(t.size()-(1+nt))) / (float)nt) * tt;
                }else {
                    u1 = tt + ((float)(i-nt) / (float)(t.size()-nt*2))* (1.0f-2.0f*tt);
                    u2 = tt + (float)(i+1-nt) / (float)(t.size()-nt*2) * (1.0f-2.0f*tt);
                }
            }else {
                u1 = (float)i / ((float)t.size()-1);
                u2 = (float)(i+1) / ((float)t.size()-1);
            }

            //Assign Line Direction to vertices
            QVector3D dir12 = (p2-p1).normalized();
            QVector3D dir01;
            if (i > 0) {
				QVector3D p0 = t[i-1];
                dir01 = (p1-p0).normalized();
            }else {
                dir01 = dir12;
            }
            QVector3D dir23;
            if (i < t.size() - 2) {
				QVector3D p3 = t[i+2];
                dir23 = (p3-p2).normalized();
            } else {
                dir23 = dir12;
            }
            QVector3D dir1 = ((dir01+dir12) / 2.0).normalized();
            QVector3D dir2 = ((dir12+dir23) / 2.0).normalized();

            //Preprocessing - Duplicated Vertices

            //Debug
            float xOffset = 0.0f;

            vertices.append(p1.x()); //Position
			vertices.append(p1.y());
			vertices.append(p1.z());
            UVs.append(u1);  //UVs
            UVs.append(0);
            directions.append(dir1.x()); //Line Direction
            directions.append(dir1.y());
            directions.append(dir1.z());

            vertices.append(p2.x());
            vertices.append(p2.y());
            vertices.append(p2.z());
            UVs.append(u2);
            UVs.append(0);
            directions.append(dir2.x());
            directions.append(dir2.y());
            directions.append(dir2.z());

            vertices.append(p1.x()+xOffset);
            vertices.append(p1.y());
            vertices.append(p1.z());
            UVs.append(u1);
            UVs.append(1);
            directions.append(dir1.x());
            directions.append(dir1.y());
            directions.append(dir1.z());

            vertices.append(p1.x()+xOffset);
            vertices.append(p1.y());
            vertices.append(p1.z());
            UVs.append(u1);
            UVs.append(1);
            directions.append(dir1.x());
            directions.append(dir1.y());
            directions.append(dir1.z());

            vertices.append(p2.x());
            vertices.append(p2.y());
            vertices.append(p2.z());
            UVs.append(u2);
            UVs.append(0);
            directions.append(dir2.x());
            directions.append(dir2.y());
            directions.append(dir2.z());

            vertices.append(p2.x()+xOffset);
            vertices.append(p2.y());
            vertices.append(p2.z());
            UVs.append(u2);
            UVs.append(1);
            directions.append(dir2.x());
            directions.append(dir2.y());
            directions.append(dir2.z());
		}
	}

	// Upload to VBO
	vbo.bind();
    vbo.allocate(vertices.data(), 9 * numVertices * sizeof(float));

    uvBuffer.bind();
	uvBuffer.allocate(UVs.data(), 6 * numVertices * sizeof(float));

    dirBuffer.bind();
    dirBuffer.allocate(directions.data(), 9 * numVertices * sizeof(float));
}

void GLWidget::resizeGL(int w, int h)
{
	float aspectRatio = static_cast<float>(w) / static_cast<float>(h);
	camera->setAspectRatio(aspectRatio);
}

void GLWidget::updateFrame()
{
	// Compute delta time
	static qint64 last = 0;
	qint64 current = frameTimer.nsecsElapsed();

	float dt = (current - last) / static_cast<float>(std::pow(1000, 3));
	(void) dt;

	last = current;

	camera->getViewMatrix(matViewModel);
	camera->getProjectionMatrix(matProj);
}

void GLWidget::paintGL()
{
	frameCounter++;

	// Wireframe mode
	if (wireframe)
	{
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	}

	// Draw
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if (program != Q_NULLPTR && numVertices > 0)
	{
		program->bind();
        QOpenGLVertexArrayObject::Binder vaoBinder(&vao);

		program->setUniformValue("m_viewModel", matViewModel);
		program->setUniformValue("m_projection", matProj);

        //Visualization parameters
		program->setUniformValue("f_thickness", interpolateSliderValue(thickness, 0.001f, 0.25f));
		program->setUniformValue("f_halo", interpolateSliderValue(halo, 0.0f, 0.25f));
		program->setUniformValue("f_stripDepth", interpolateSliderValue(stripDepth, 0.0f, 1.0f));
		program->setUniformValue("f_depthCueing", interpolateSliderValue(depthCueing, 0.0f, 1.0f));
        //---------

		textureStroke->bind(0);
        textureStripMask->bind(1);

        glDrawArrays(GL_TRIANGLES, 0, numVertices*3);

		textureStroke->release(0);
        textureStripMask->release(1);
		program->release();
	}

	// Reset polygon mode
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}

void GLWidget::keyPressEvent(QKeyEvent* event)
{
	if (event->key() == Qt::Key_F && !event->isAutoRepeat())
	{
		filterTracks = !filterTracks;
		setTrackData(std::move(tracks));
	}
}

void GLWidget::mousePressEvent(QMouseEvent *event)
{
    m_lastPos = event->pos();
}

void GLWidget::mouseMoveEvent(QMouseEvent *event)
{
    int dx = event->x() - m_lastPos.x();
    int dy = event->y() - m_lastPos.y();

    if (event->buttons() & Qt::LeftButton) {
		camera->rotateX(0.01f * dx);
		camera->rotateY(-0.01f * dy);
    } else if (event->buttons() & Qt::RightButton) {
		camera->zoom(dy);
    } else if (event->buttons() & Qt::MidButton) {
		camera->move(QVector2D(0.01f*dx, 0.01f*-dy));
    }
    m_lastPos = event->pos();
}

void GLWidget::wheelEvent(QWheelEvent *event)
{
	camera->zoom(-event->angleDelta().y() * 0.05f);
}

float GLWidget::interpolateSliderValue(int value, float min, float max) {
    return min + (((float)value / 100.0f) * (max-min));
}

QVector3D GLWidget::normalizedPoint(QVector3D point, QVector3D min, QVector3D range, float scale) {
    //X = ((X - (minX+(maxRange)*0.5)) * scale);
    QVector3D npoint = (point - (min + range*0.5) ) * scale;
    return npoint;
}
