#ifndef GLWIDGET_H
#define GLWIDGET_H

#include "trackfile.h"
#include "camera.h"

#include <QOpenGLWidget>
#include <QOpenGLFunctions_4_0_Core>
#include <QOpenGLDebugLogger>
#include <QTimer>
#include <QElapsedTimer>
#include <QOpenGLShaderProgram>
#include <QOpenGLBuffer>
#include <QOpenGLVertexArrayObject>
#include <QMatrix4x4>
#include <QOpenGLTexture>

/**
 * @brief Widget encapsulating the OpenGL context.
 */
class GLWidget : public QOpenGLWidget, QOpenGLFunctions_4_0_Core
{
	Q_OBJECT

	public:
		/**
		 * @brief Constructs a new OpenGL widget.
		 * @param parent the parent widget.
		 * @param updateRate the update rate of the widget.
		 * @param sampleShading indicates if sample shading is to be used.
		 * @param anisotropy the maximum number of samples used for anisotropic filtering.
		 */
		GLWidget(QWidget* parent = Q_NULLPTR, int updateRate = 60, bool sampleShading = true, int anisotropy = 1);

		~GLWidget();

		/**
		 * @brief Prepares track data for rendering.
		 * @param tracks the track data.
		 */
		virtual void setTrackData(const QVector<Track>&& tracks);

	public slots:

		/**
		 * @brief Enables / disables wireframe mode.
		 * @param enabled true for wireframe rendering, false for solid shading.
		 */
		virtual void setWireframe(bool enabled)
		{
			wireframe = enabled;
		}

        /**
         * @brief Sets the line thickness value
         * @param p_thickness the value
         */
        virtual void setThickness(int p_thickness) {
            thickness = p_thickness;
        }

        /**
         * @brief Sets the halo width value
         * @param p_halo the value
         */
        virtual void setHalo(int p_halo) {
            halo = p_halo;
        }

        /**
         * @brief Sets the strip depth value
         * @param p_stripDepth the value
         */
        virtual void setStripDepth(int p_stripDepth){
            stripDepth = p_stripDepth;
        }

        /**
         * @brief sets the depth cueing value
         * @param p_depthCueing the value
         */
        virtual void setDepthCueing(int p_depthCueing){
            depthCueing = p_depthCueing;
        }

	protected slots:

		/**
		 * @brief Updates the scene.
		 *
		 * This function is called at a fixed interval, which is determined in the constructor.
		 * Here everything related to updating the scene should
		 */
		virtual void updateFrame();

		/**
		 * @brief Renders the scene.
		 *
		 * This function is called whenever the widget needs to be updated (e.g. resizing), as
		 * well as at a fixed interval.
		 */
		virtual void paintGL() Q_DECL_OVERRIDE;

		/**
		 * @brief Initializes everything related to OpenGL.
		 *
		 * This function is called after the context has been created, before the first invocation of
		 * initializeGL() and resizeGL()
		 */
		virtual void initializeGL() Q_DECL_OVERRIDE;

		/**
		 * @brief Handles resizing of the widget.
		 * @param w the new width.
		 * @param h the new height.
		 *
		 * This function is called whenever the widget is resized (e.g. user input).
		 */
		virtual void resizeGL(int w, int h) Q_DECL_OVERRIDE;

	signals:

		/**
		 * @brief Signals the current frames per second periodically.
		 * @param fps the current frames per second.
		 */
		void frameRate(int fps);

    protected:

		/**
		 * @brief Handles key presses for toggling track filtering.
		 * @param event the key event.
		 */
		void keyPressEvent(QKeyEvent* event) Q_DECL_OVERRIDE;

        /**
         * @brief Handles mouse click interaction for camera movement
         * @param event the mouse event
         */
        void mousePressEvent(QMouseEvent *event) override;

        /**
         * @brief Handles mouse movement interaction for camera movement
         * @param event the mouse event
         */
        void mouseMoveEvent(QMouseEvent *event) override;

		/**
		 * @brief Handles mouse wheel interaction for camera movement.
		 * @param event the wheel event.
		 */
		void wheelEvent(QWheelEvent* event) Q_DECL_OVERRIDE;

	private:

		/**
		 * @brief Creates a shader program given a vertex, fragment and (optionally) geometry shader.
		 * @param pathVS the path to the vertex shader file.
		 * @param pathFS the path to the fragment shader file.
		 * @param pathGS the path to the geometry shader file, or an empty string.
		 * @return the linked shader program or Q_NULLPTR if an error occurred.
		 */
		QOpenGLShaderProgram* createShaderProgram(const QString& pathVS, const QString& pathFS, const QString& pathGS = "");

		// Logger
		QOpenGLDebugLogger* logger = Q_NULLPTR;

		// Counter for frames
		int frameCounter = 0;

		// Called every second to compute fps
		QTimer fpsTimer;

		// Updates the widget periodically, to force paintGL() invocation
		// Also calls updateFrame()
		QTimer updateTimer;

		// Timer to compute delta time
		QElapsedTimer frameTimer;

		// Flag for sample shading
		const bool sampleShading;

		// Maximum number of samples used for anisotropic filtering
		const int anisotropy;

		// Flag for wireframe mode
		bool wireframe = false;

        // Line thickness value
		int thickness = 0;

        // Halo size value
		int halo = 0;

        // Depth cueing value
        int depthCueing = 0;

        // Strip depth value
		int stripDepth = 0;

		// Indicates if loaded tracks should be filtered by just considering
		// tracks that intersect with z = 0
		bool filterTracks = false;

		// The current track data
		QVector<Track> tracks;

		// Texture for the strips
		QOpenGLTexture* textureStroke = Q_NULLPTR;

        // Texture for the tapering stencil mask
        QOpenGLTexture* textureStripMask = Q_NULLPTR;

		// Program for rendering
		QOpenGLShaderProgram* program = Q_NULLPTR;

		// Vertex buffer
		QOpenGLBuffer vbo;

        // UV buffer
        QOpenGLBuffer uvBuffer;

        // Directions buffer
        QOpenGLBuffer dirBuffer;

		// Vertex array object
		QOpenGLVertexArrayObject vao;

		// Camera
		Camera* camera = nullptr;

		// View model matrix
		QMatrix4x4 matViewModel;

        // Proj matrix
        QMatrix4x4 matProj;

		// Number of vertices (for current dataset)
		int numVertices = 0;

        // Last mouse position (for camera movement)
        QPoint m_lastPos;

        /**
         * @brief Interpolates the integer value from the slider to the specified range
         * @param value the value from the slider from 0 to 99;
         * @param min the minimum value
         * @param max the maximum value
		 * @return the interpolated value
         */
        float interpolateSliderValue(int value, float min, float max);

        /**
         * @brief returns the normalized position for this point
         * @param point the point
         * @param min the global minimum
         * @param range the maximum range of the 3 dimensions
		 * @return the normalized position
         */
        QVector3D normalizedPoint(QVector3D point, QVector3D min, QVector3D range, float scale);
};

#endif // GLWIDGET_H
