#include "mainwindow.h"
#include "ui_mainwindow.h"

#include <QDebug>
#include <QFileDialog>
#include <QOffscreenSurface>
#include <QOpenGLFunctions>
#include <QSettings>

MainWindow::MainWindow(QWidget *parent) :
	QMainWindow(parent),
	ui(new Ui::MainWindow)
{
	ui->setupUi(this);

	// Setup signals / slots
	connect(ui->actionQuit, &QAction::triggered, this, &MainWindow::close);
	connect(ui->actionLoad, &QAction::triggered, this, &MainWindow::load);

	// Read settings
	int maxSamples = getMaximumSamples();

	QSettings settings("settings.ini", QSettings::IniFormat);
	int samples = qMin(maxSamples, settings.value("Settings/msaa", maxSamples).toInt());
	bool sampleShading = settings.value("Settings/sampleshading", true).toBool();
	int anisotropy = settings.value("Settings/anisotropy", 1).toInt();

	// Create GL widget
	glWidget = new GLWidget(this, 60, sampleShading, anisotropy);

	QSurfaceFormat format;
	format.setVersion(4, 5);
	format.setProfile(QSurfaceFormat::CoreProfile);
	format.setOption(QSurfaceFormat::DebugContext);
	format.setSamples(samples);
	format.setSwapInterval(0);

	glWidget->setFormat(format);
	glWidget->setMinimumSize(640, 480);

	// Add GL widget to main grid layout
	QGridLayout* l = dynamic_cast<QGridLayout*>(ui->centralWidget->layout());
	l->addWidget(glWidget, 1, 0, 1, 1);

	// Connect LCD display to fps signal of GL widget
	connect(glWidget, &GLWidget::frameRate, [this] (int fps) {
		ui->lcdFPS->display(fps);
	});

	ui->lcdFPS->setPalette(Qt::darkGreen);

	// Wireframe toggle
	connect(ui->checkBoxWireframe, &QCheckBox::toggled, glWidget, &GLWidget::setWireframe);

    // Sliders
    connect(ui->sliderThickness, &QSlider::valueChanged, glWidget, &GLWidget::setThickness);
    connect(ui->sliderHalo, &QSlider::valueChanged, glWidget, &GLWidget::setHalo);
    connect(ui->sliderStripDepth, &QSlider::valueChanged, glWidget, &GLWidget::setStripDepth);
    connect(ui->sliderDepthCueing, &QSlider::valueChanged, glWidget, &GLWidget::setDepthCueing);

	glWidget->setThickness(ui->sliderThickness->value());
	glWidget->setHalo(ui->sliderHalo->value());
	glWidget->setStripDepth(ui->sliderStripDepth->value());
	glWidget->setDepthCueing(ui->sliderDepthCueing->value());
}

MainWindow::~MainWindow()
{
	delete ui;
}

void MainWindow::load()
{
	// Prompt user to open track file
	QString fname = QFileDialog::getOpenFileName(this, "Open data file", "./data", "Track files (*.trk)");

	// Read data
	if (!fname.isNull())
	{
		QVector<Track> tracks;

		if (TrackFile::read(fname, tracks))
		{
			glWidget->setTrackData(std::move(tracks));
			ui->labelStatus->setText(QString("Loaded data '%1'").arg(fname));
		}
		else
		{
			ui->labelStatus->setText(QString("Error reading file '%1'").arg(fname));
		}
	}
}

int MainWindow::getMaximumSamples()
{
	QOpenGLContext ctx;
	QOffscreenSurface surface;
	GLint maxSamples = 1;

	if (ctx.create())
	{
		surface.setFormat(ctx.format());
		surface.create();

		ctx.makeCurrent(&surface);
		ctx.functions()->glGetIntegerv(GL_MAX_SAMPLES, &maxSamples);
		ctx.doneCurrent();
	}

	return maxSamples;
}
