#include "trackfile.h"

#include <cstdint>
#include <QDebug>
#include <QDataStream>
#include <QFile>

bool TrackFile::read(const QString &fname, QVector<Track>& tracks)
{
	// Open file
	QFile f(fname);

	if (!f.open(QFile::ReadOnly))
	{
		qDebug("Failed to open file '%s'", qPrintable(fname));
		return false;
	}

	// Setup stream
	QDataStream is(&f);

	// Read and check ID
	char id[6];
	is.readRawData(id, 6);
	id[5] = '\0';			// Doc does not specify that it is 0-terminated

	if (!QString(id).startsWith("TRACK"))
	{
		qDebug("File '%s' has unknown format (ID mismatch)", qPrintable(fname));
		return false;
	}

	is.skipRawData(6);			// dim
	is.skipRawData(12);			// voxel_size
	is.skipRawData(12);			// origin (always 0,0,0)

	// Read number of scalars
	int16_t numScalars;
	is.readRawData(reinterpret_cast<char*>(&numScalars), 2);

	is.skipRawData(200);		// scalar_name

	// Read number of properties
	int16_t numProperties;
	is.readRawData(reinterpret_cast<char*>(&numProperties), 2);

	is.skipRawData(200);		// property_name
	is.skipRawData(64);			// vox_to_ras
	is.skipRawData(444);		// reserved
	is.skipRawData(4);			// voxel_order
	is.skipRawData(4);			// pad2
	is.skipRawData(24);			// image_orientation_patient
	is.skipRawData(2);			// pad1
	is.skipRawData(1);			// invert_x
	is.skipRawData(1);			// invert_y
	is.skipRawData(1);			// invert_z
	is.skipRawData(1);			// swap_xy
	is.skipRawData(1);			// swap_yz
	is.skipRawData(1);			// swap_zx

	// Read number of tracks
	int32_t numTracks;
	is.readRawData(reinterpret_cast<char*>(&numTracks), 4);

	is.skipRawData(4);			// version
	is.skipRawData(4);			// hdr_size

	// Read track data
	tracks.clear();

	if (numTracks > 0)
	{
		tracks.reserve(numTracks);
	}

    QVector3D prevPoint;
    float th = 0.125f;
    int iskipped = 0;

	while (!is.atEnd() && is.status() == QDataStream::Ok)
	{
		Track trk;

		// Number of points in current track
		int32_t numPoints;
		is.readRawData(reinterpret_cast<char*>(&numPoints), 4);

		// Read points
		for (int i = 0; i < numPoints; i++)
		{
			float p[3];
			is.readRawData(reinterpret_cast<char*>(&p), 12);
			is.skipRawData(numScalars * 4);		// scalars

            QVector3D point = QVector3D(p[0], p[2], p[1]);
            float dist = point.distanceToPoint(prevPoint);

            if (i == 0) {
                prevPoint = point;
            }
            else if (dist > th) {
                trk.append(point);
                prevPoint = point;
            }else {
                iskipped++;
                //printf("Skipped point. Distance = %f\n", dist);
            }
		}

		is.skipRawData(numProperties * 4);	// properties

		if (numPoints > 1)
		{
			tracks.append(trk);
		}
		else
		{
			numTracks--;
			qDebug("Ignored track with %d points", numPoints);
		}
	}

    printf("skipped points = %d\n", iskipped);

	// Do we have all tracks?
	if (numTracks > 0 && tracks.size() != numTracks)
	{
		qDebug("Read %d tracks, but expected %d", tracks.size(), numTracks);
		return false;
	}

	// Return
	return is.status() == QDataStream::Ok;
}
